/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamsongdatabase.h"
#include "gdamutil.h"
#include <gdam/gdamtildeexpand.h>
#include <ctype.h>
#include "../common/gdamdebug.h"
#include <string.h>
#include <sys/stat.h>

enum {
	CHANGED,
	LAST_SIGNAL
};

static guint database_signals[LAST_SIGNAL] = { 0 };


static void gdam_song_database_destroy(GtkObject* object);

static GdamSongDatabase* default_song_database = NULL;
static GtkObjectClass* parent_class = NULL;

static int compare_by_pointer(gpointer a, gpointer b)
{
	if (GPOINTER_TO_INT(a) < GPOINTER_TO_INT(b))
		return -1;
	if (GPOINTER_TO_INT(a) > GPOINTER_TO_INT(b))
		return +1;
	return 0;
}

static int gdam_song_info_by_order(GdamSongInfo* a, GdamSongInfo* b)
{
	int aorder = gdam_song_info_get_order(a);
	int border = gdam_song_info_get_order(b);
	if (aorder < border)
		return -1;
	if (aorder > border)
		return +1;
	return compare_by_pointer(a, b);
}

static int gdam_song_info_by_name(GdamSongInfo* a, GdamSongInfo* b)
{
	const char* afilename = gdam_song_info_get_filename(a);
	const char* bfilename = gdam_song_info_get_filename(b);
	int order = strcmp(afilename, bfilename);
	if (order == 0)
		return compare_by_pointer(a, b);
	return order;
}
static int gdam_song_info_by_length(GdamSongInfo* a, GdamSongInfo* b)
{
	float length_a = gdam_song_info_get_length(a);
	float length_b = gdam_song_info_get_length(b);
	if (length_a < 0) length_a = 1000000;
	if (length_b < 0) length_b = 1000000;
	if (length_a < length_b)
		return -1;
	if (length_a > length_b)
		return +1;
	return compare_by_pointer(a, b);
}

static int gdam_song_info_by_bpm(GdamSongInfo* a, GdamSongInfo* b)
{
	/* TODO: need to free memory from si_get_bi? */
	float abpm = 
		gdam_beat_info_get_bpm(gdam_song_info_get_beat_info(a),0);
	float bbpm = 
		gdam_beat_info_get_bpm(gdam_song_info_get_beat_info(b),0);
	if (abpm < bbpm)
		return -1;
	if (abpm > bbpm)
		return +1;
	return compare_by_pointer(a, b);
}

GdamSongDatabase*    gdam_song_database_new(const char* filename)
{
	GdamSongDatabase* database;
	database = GDAM_SONG_DATABASE(gtk_type_new(GDAM_TYPE_SONG_DATABASE));
	database->dbname = g_strdup(filename);
	return database;
}


void                 gdam_song_database_add     (GdamSongDatabase* db,
                                                 GdamSongInfo*     song_info)
{
	const char* filename = gdam_song_info_get_filename(song_info);
	if (g_hash_table_lookup(db->name_lookup, filename))
		gdam_song_database_remove(db, filename);
	g_hash_table_insert(db->name_lookup, g_strdup(filename), song_info);
	g_tree_insert(db->by_order, song_info, song_info);
	g_tree_insert(db->by_name, song_info, song_info);
	g_tree_insert(db->by_length, song_info, song_info);
	g_tree_insert(db->by_bpm, song_info, song_info);
}

GdamSongInfo*    gdam_song_database_find(GdamSongDatabase* database,
					 const char* filename)
{
	return g_hash_table_lookup(database->name_lookup, filename);
	
}

void                 gdam_song_database_remove  (GdamSongDatabase* db,
                                                 const char*       filename)
{
	GdamSongInfo* to_kill;
	const char* doomed_fname;
	to_kill = (GdamSongInfo*)g_hash_table_lookup(db->name_lookup,filename);
	doomed_fname = gdam_song_info_get_filename(to_kill);
	if (!to_kill)
		return;
	g_hash_table_remove(db->name_lookup, doomed_fname);
	g_tree_remove(db->by_order, to_kill);
	g_tree_remove(db->by_name, to_kill);
	g_tree_remove(db->by_length, to_kill);
	g_tree_remove(db->by_bpm, to_kill);
	gdam_song_info_destroy(to_kill);
}

static void kill_songinfo(gpointer key, gpointer value, gpointer unused)
{
	gdam_song_info_destroy((GdamSongInfo*)value);
}
static void print_props(gpointer key, gpointer value, gpointer data)
{
	FILE* fp = (FILE*)data;
	fprintf(fp, "%s: %s\n", (char*)key, (char*)value);
}

static gint gdam_song_info_print(gpointer key, gpointer value, gpointer data)
{
	GdamSongInfo* song_info = (GdamSongInfo*)value;
	/* TODO: free beat_info? */
	GdamBeatInfo* beat_info = gdam_song_info_get_beat_info(song_info);
	float first_beat = gdam_beat_info_get_first_beat(beat_info,0);
	float bpm        = gdam_beat_info_get_bpm(beat_info,0);
	float length     = gdam_song_info_get_length(song_info);
	float gain       = gdam_song_info_get_gain(song_info);
	const char* filename = gdam_song_info_get_filename(song_info);
	FILE* fp = (FILE*)data;
	const GHashTable* prop_table;
	fprintf(fp, "file: %s\n", filename);
	if (first_beat >= 0.0)
		fprintf(fp, "firstbeat: %3.6f\n", first_beat);
	if (length >= 0.0)
		fprintf(fp, "length: %3.6f\n", length);
	if (bpm >= 0.0)
		fprintf(fp, "bpm: %3.6f\n", bpm);
	if (gain >= 0.0)
		fprintf(fp, "gain: %3.6f\n", gain);

	prop_table = gdam_song_info_get_properties(song_info);
	if (prop_table){
		g_hash_table_foreach((GHashTable*)prop_table, 
				     print_props, 
				     fp);
	}

	fprintf(fp, "\n");
	return 0;
}

void                 gdam_song_database_write_fp(GdamSongDatabase* db,
                                                 FILE* fp)
{
	g_tree_traverse(db->by_order, 
			(GTraverseFunc) gdam_song_info_print,
	                G_IN_ORDER,
			fp);
}

void                 gdam_song_database_write   (GdamSongDatabase* db,
                                                 const char* fname)
{
	FILE* fp;
	char* expanded_filename = gdam_tilde_expand(fname);
	fp = fopen(expanded_filename, "w");
	g_free(expanded_filename);
	if (!fp) {
		fprintf(stderr, "couldn't open %s for writing.\n", fname);
		return;
	}
	gdam_song_database_write_fp(db, fp);
	fclose(fp);
}


GdamSongDatabase* 
gdam_song_database_new_from_file(const char* filename)
{
	char buf[2048];
	FILE* fp;
	int lineno = 0;
	GdamSongDatabase* database;
	GdamSongInfo* song_info = NULL;
	int order = 1;
	fp = fopen(filename, "r");
	if (!fp) {
		g_warning("couldn't open `%s' for reading", filename);
		return NULL;
	}
	
	database = gdam_song_database_new(filename);
	while (fgets(buf, sizeof(buf), fp)) {
		char* key = buf;
		char* colon; char* value;
		lineno++;
		while (*key && isspace(*key))
			key++;
		if (*key == '#' || *key == '\0') continue;
		colon = strchr(buf, ':');
		if (!colon) {
			g_warning("parsing %s, line %d: no colon",
			    	filename, lineno);
			continue;
		}
		*colon = 0;
		value = colon + 1;
		while (*value && isspace(*value))
			value++;
		g_strchomp(value);
		g_strdown(key);
		/*
		gdam_debug("song_database_parse: key=%s; value=%s\n",
				key, value);
		*/
		if (strcmp(key, "file") == 0) {
			if (song_info)
				gdam_song_database_add(database, song_info);
			song_info = gdam_song_info_new(value, 0.0, 0.0);
		} else {
			if (!song_info) {
				g_warning("first key of %s was `%s', must be "
					"`file' (line %d)",
					filename, key, lineno);
				continue;
			}
			gdam_song_info_add_property(song_info, key, value);
		}
	}
	if (song_info) {
		gdam_song_info_set_order(song_info, order++);
		gdam_song_database_add(database, song_info);
	} else
		g_warning("%s had no entries", filename);
	return database;
}

static const char* fnames[] = {
	"/etc/gdam/song.database",
	"~/.gdam/song.database",
	(const char*)NULL
};

const char* gdam_song_database_default_filename = NULL;

GdamSongDatabase*
gdam_song_database_system()
{
	const char* const* at;
	GdamSongDatabase* db;
	gdam_debug("Song Database System called");
	if (default_song_database){
		gdam_debug("Song Database System..  returning default");
		return default_song_database;
	}
	if (gdam_song_database_default_filename != NULL) {
		db = gdam_song_database_new_from_file(
			gdam_song_database_default_filename
		);
		if (db == NULL) {
			g_error("song database %s does not exist",
				gdam_song_database_default_filename);
		}
		return db;
	}
#if HAVE_GETENV
	{
		char* db_env;
		db_env = getenv("GDAM_SONG_DATABASE");
		if (db_env != NULL) {
			db = gdam_song_database_new_from_file(db_env);
			if (db == NULL) {
				g_warning("error loading song database %s",
					db_env);
			} else {
				return db;
			}
		}
	}
#endif
	for (at = fnames; *at; at++) {
		char* name = gdam_tilde_expand(*at);
		if (name) {
			if (gdam_file_exists(name)) {
				db = gdam_song_database_new_from_file(name);
				gtk_object_ref(GTK_OBJECT(db));
				gdam_debug("Song Database from %s", name);
				return default_song_database = db;
			}
			g_free(name);
		}
	}
	db = gdam_song_database_new("system");
	gdam_debug("Creating new database");
	return default_song_database = db;
}


static void gdam_song_database_init(GdamSongDatabase* song_database)
{
	song_database->by_order
		= g_tree_new((GCompareFunc)gdam_song_info_by_order);
	song_database->by_name
		= g_tree_new((GCompareFunc)gdam_song_info_by_name);
	song_database->by_length
		= g_tree_new((GCompareFunc)gdam_song_info_by_length);
	song_database->by_bpm
		= g_tree_new((GCompareFunc)gdam_song_info_by_bpm);
	song_database->name_lookup
		= g_hash_table_new(g_str_hash, g_str_equal);
}
static void gdam_song_database_class_init(GdamSongDatabaseClass* klass)
{
	GtkObjectClass* object_class = GTK_OBJECT_CLASS(klass);
	object_class->destroy = gdam_song_database_destroy;
	database_signals[CHANGED]
		= gtk_signal_new("changed",
		                 GTK_RUN_LAST,
				object_class->type,
				GTK_SIGNAL_OFFSET(GdamSongDatabaseClass,
						  changed),
				gtk_marshal_NONE__NONE,
				GTK_TYPE_NONE,
				0);
	gtk_object_class_add_signals(object_class,
	                             database_signals,
				     LAST_SIGNAL);
}

GtkType gdam_song_database_get_type()
{
	static GtkType song_database_type = 0;
	if (!song_database_type) {
		static const GtkTypeInfo song_database_info = {
			"GdamSongDatabase",
			sizeof(GdamSongDatabase),
			sizeof(GdamSongDatabaseClass),
			(GtkClassInitFunc) gdam_song_database_class_init,
			(GtkObjectInitFunc) gdam_song_database_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GTK_TYPE_OBJECT;
		parent_class = gtk_type_class(parent);
		song_database_type = gtk_type_unique(parent,
		                                 &song_database_info);
	}
	return song_database_type;
}

static void gdam_song_database_destroy(GtkObject* object)
{
	GdamSongDatabase* database = GDAM_SONG_DATABASE(object);
	g_tree_destroy(database->by_order);
	g_tree_destroy(database->by_name);
	g_tree_destroy(database->by_length);
	g_tree_destroy(database->by_bpm);
	g_hash_table_foreach(database->name_lookup, kill_songinfo, NULL);
	g_hash_table_destroy(database->name_lookup);
	(*parent_class->destroy)(object);
}
