/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamsource.h"
#include "gdamdispatch.h"
#include <gdam/gdamplugin.h>
#include <gdam/gdam-protocol.h>
#include <gtk/gtksignal.h>
#include <ctype.h>
#include "gdamsource.h"
#include "gdamconnection.h"
#include <gdam/gdammodule.h>
#include <gdam/gdamnamespace.h>
#include <stdio.h>
#include "gdamevent.h"


/* Hardcore debugging of missed data. */
#define GDAM_DEBUG_MISS_DATA 1

static void gdam_source_class_init       (GdamSourceClass*  klass);
static void gdam_source_init             (GdamSource*       source);
static void gdam_source_shutdown         (GdamSource*       source);
static void gdam_source_destroy          (GdamSource*       source);
static void gdam_source_ended            (GdamSource*       source);
static void gdam_source_get_latency_info (GdamSource*       source,
                                          GdamSource*       subsource,
					  GdamLatencyInfo*  latency);
static GdamTime
            gdam_source_real_get_position(GdamSource*       source);
static GdamBuffer*
            gdam_source_real_interpret_command
                                         (GdamSource*       source,
	                                  GdamBuffer*       command,
					  gboolean*         was_success);
#if 0
static void gdam_source_set_arg          (GtkObject*        source,
                                          GtkArg*           arg,
					  guint             arg_id);
static void gdam_source_get_arg          (GtkObject*        source,
                                          GtkArg*           arg,
					  guint             arg_id);
#endif

static GtkObjectClass* the_object_class = NULL;

GtkType gdam_source_get_type()
{
	static GtkType source_type = 0;
	if (!source_type) {
		static const GtkTypeInfo source_info = {
			"GdamSource",
			sizeof(GdamSource),
			sizeof(GdamSourceClass),
			(GtkClassInitFunc) gdam_source_class_init,
			(GtkObjectInitFunc) gdam_source_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		the_object_class = gtk_type_class(GTK_TYPE_OBJECT);
		source_type = gtk_type_unique(GTK_TYPE_OBJECT, &source_info);
	}
	return source_type;
}

static void gdam_source_arg_set_position(GtkObject* object, GdamArg* arg)
{
	gdam_source_seek(GDAM_SOURCE(object), GDAM_ARG_INT64(arg), 0);
}

static void gdam_source_arg_get_position(GtkObject* object, GdamArg* arg)
{
	GDAM_ARG_INT64(arg) = gdam_source_get_position(GDAM_SOURCE(object));
}

static void gdam_source_arg_set_synchronize(GtkObject* object, GdamArg* arg)
{
	int add_id, fixed_id;
	gint64 mobile_period, mobile_offset;
	gint64 fixed_period, fixed_offset;
	GdamSource* mobile = GDAM_SOURCE(object);
	GdamSource* fixed;
	GdamSource* mobile_latency_proxy;
	int seek;

	g_return_if_fail(GDAM_ARG_INT64_ARRAY(arg).length == 6);

	add_id        = (int) GDAM_ARG_INT64_ARRAY(arg).array[0];
	fixed_id      = (int) GDAM_ARG_INT64_ARRAY(arg).array[1];
	mobile_period =       GDAM_ARG_INT64_ARRAY(arg).array[2];
	mobile_offset =       GDAM_ARG_INT64_ARRAY(arg).array[3];
	fixed_period  =       GDAM_ARG_INT64_ARRAY(arg).array[4];
	fixed_offset  =       GDAM_ARG_INT64_ARRAY(arg).array[5];

	if (add_id == 0)
		mobile_latency_proxy = NULL;
	else
		mobile_latency_proxy = gdam_namespace_local_source(add_id);
	fixed = gdam_namespace_local_source(fixed_id);
	if (fixed == NULL) {
		gdam_debug("fixed_id %d didn't exist", fixed_id);
		return;
	}
	seek = gdam_source_get_sync_seek(mobile, fixed,
	                                 mobile_period, mobile_offset,
	                                 fixed_period, fixed_offset,
					 mobile_latency_proxy);
	gdam_source_seek(mobile, seek, 1);

	gdam_debug("Checking sync seek now...");
	gdam_debug("would now seek %d", 
			gdam_source_get_sync_seek(mobile, fixed,
	                                 mobile_period, mobile_offset,
	                                 fixed_period, fixed_offset,
					 mobile_latency_proxy));
}

static void gdam_source_arg_get_subsources(GtkObject* object, GdamArg* arg)
{
	int num_subsources;
	GdamSource** subsources;
	int* array;
	int i;
	gdam_source_get_subsources(GDAM_SOURCE(object),
	                           &num_subsources,
				   &subsources);
	GDAM_ARG_INT_ARRAY(arg).length = num_subsources;
	GDAM_ARG_INT_ARRAY(arg).array = array = g_new(int, num_subsources);
	for (i = 0; i < num_subsources; i++)
		array[i] = subsources[i]->source_id;
}

static void gdam_source_arg_get_parent(GtkObject* object, GdamArg* arg)
{
	GdamSource* parent = GDAM_SOURCE(object)->parent;
	GDAM_ARG_INT(arg) = parent ? parent->source_id : 0;
}
static void gdam_source_arg_get_parent_device(GtkObject* object, GdamArg* arg)
{
	GdamSink* sink = GDAM_SOURCE(object)->device;
	GDAM_ARG_INT(arg) = sink ? sink->sink_id : 0;
}
static void gdam_source_arg_get_paused(GtkObject* object, GdamArg* arg)
{
	GDAM_ARG_INT(arg) = GDAM_SOURCE(object)->paused;
}
static void gdam_source_arg_set_paused(GtkObject* object, GdamArg* arg)
{
	GDAM_SOURCE(object)->paused = GDAM_ARG_INT(arg) ? 1 : 0;
}

enum _SourceSignals {
	ENDED,
	LAST_SIGNAL
} SourceSignals;

static guint source_signals[LAST_SIGNAL];

static void gdam_source_class_init(GdamSourceClass* klass)
{
	GtkObjectClass* object_class = (GtkObjectClass*)klass;
	klass->interpret_command = gdam_source_real_interpret_command;
	klass->ended = gdam_source_ended;
	klass->get_position = gdam_source_real_get_position;
	klass->get_latency_info = gdam_source_get_latency_info;
	source_signals[ENDED] =
		gtk_signal_new("ended",
			       GTK_RUN_FIRST,
			       object_class->type,
			       GTK_SIGNAL_OFFSET(GdamSourceClass, ended),
			       gtk_marshal_NONE__NONE,
			       GTK_TYPE_NONE, 0);
	object_class->destroy = (void (*)(GtkObject*)) gdam_source_destroy;
	object_class->shutdown = (void (*)(GtkObject*)) gdam_source_shutdown;
	gtk_object_class_add_signals(object_class, source_signals, LAST_SIGNAL);

	gdam_arg_register(object_class,
	                  "position",
			  GDAM_ARG_TYPE_INT64,
			  gdam_source_arg_set_position,
			  gdam_source_arg_get_position,
			  0);
	gdam_arg_register(object_class,
	                  "synchronize",
			  GDAM_ARG_TYPE_INT64_ARRAY,
			  gdam_source_arg_set_synchronize,
			  NULL,
			  0);
	gdam_arg_register(object_class,
	                  "subsources",
			  GDAM_ARG_TYPE_INT_ARRAY,
			  NULL,
			  gdam_source_arg_get_subsources,
			  0);
	gdam_arg_register(object_class,
	                  "paused",
			  GDAM_ARG_TYPE_INT,
			  gdam_source_arg_set_paused,
			  gdam_source_arg_get_paused,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "parent",
			  GDAM_ARG_TYPE_INT,
			  NULL,
			  gdam_source_arg_get_parent,
			  0);
	gdam_arg_register(object_class,
	                  "parent_device",
			  GDAM_ARG_TYPE_INT,
			  NULL,
			  gdam_source_arg_get_parent_device,
			  0);
}

static void gdam_source_init(GdamSource* source)
{
	source->parent = NULL;
	source->device = NULL;
	source->source_id = gdam_dispatch_obtain_source_id(source);
}

static void gdam_source_shutdown(GdamSource* source)
{
	if (source->source_id) {
		gdam_debug("releasing source id %d", source->source_id);
		gdam_dispatch_release_source_id(source->source_id);
	}
	the_object_class->shutdown((GtkObject*)source);
}
static void gdam_source_destroy(GdamSource* source)
{
	if (source->parent || source->device) {
		gdam_source_orphan(source);
		gtk_object_unref((GtkObject*)source);
	}
	gdam_debug("gdam_source_destroy: %p; id=%d", source, source->source_id);
	g_assert(gdam_dispatch_get_source(source->source_id) == NULL);
	the_object_class->destroy((GtkObject*)source);
}
static void gdam_source_get_latency_info (GdamSource*       source,
                                          GdamSource*       subsource,
					  GdamLatencyInfo*  latency)
{
	latency->latency = 0;
	latency->resample = GdamFixed_ONE;
}

int gdam_source_read(GdamSource* input, GdamSample* samples, int max_read)
{
	GtkObjectClass* klass = ((GtkObject*)input)->klass;
	GdamSourceClass* source_class = (GdamSourceClass*)klass;
	int rv;
	if (input->paused) {
		gdam_sample_zero(samples, max_read);
		return max_read;
	}
	rv = source_class->read(input, samples, max_read);
	if (rv > max_read) {
		g_error("gdam_source_read: %s: "
		        "read %d samples when asked for %d samples",
			gtk_type_name(klass->type), rv, max_read);
	}
	if (rv == 0 && !input->ended)
		gtk_signal_emit((GtkObject*)input, source_signals[ENDED]);
	return rv;
}

#if GDAM_DEBUG_MISS_DATA
static GHashTable* missed_data_by_type = NULL;
static guint       flush_thread = 0;

static void print_type_and_count(gpointer key, gpointer value, gpointer unused)
{
	fprintf(stderr, " %s[%d]", gtk_type_name((GtkType) key),
	                           GPOINTER_TO_INT(value));
}
static gboolean flush_missed_data_hash(gpointer unused)
{
	if (g_hash_table_size(missed_data_by_type) > 0) {
		fprintf(stderr, "Missed data:");
		g_hash_table_foreach(missed_data_by_type,
		                     print_type_and_count,
				     NULL);
		fprintf(stderr, ".\n");
	}
	g_hash_table_destroy(missed_data_by_type);
	missed_data_by_type = g_hash_table_new(NULL, NULL);
	return TRUE;
}
static void gdam_debug_note_missed_data(GdamSource* source, int num)
{
	GtkType type = GTK_OBJECT(source)->klass->type;
	int amt;
	if (missed_data_by_type == NULL) {
		missed_data_by_type = g_hash_table_new(NULL, NULL);
		flush_thread = g_timeout_add(1000,
		                             flush_missed_data_hash,
					     NULL);
	}
	amt = (int) g_hash_table_lookup(missed_data_by_type, (gpointer) type);
	amt += num;
	g_hash_table_insert(missed_data_by_type,
	                    (gpointer) type,
			    (gpointer) amt);
}
#endif


void gdam_source_miss_data(GdamSource* input, int samples_lost)
{
	GtkObjectClass* klass = ((GtkObject*)input)->klass;
	GdamSourceClass* source_class = (GdamSourceClass*)klass;
	g_return_if_fail(samples_lost > 0);
#if GDAM_DEBUG_MISS_DATA
	if (gdam_debugging_on)
		gdam_debug_note_missed_data(input, samples_lost);
#endif
	source_class->miss_data(input, samples_lost);
}
gboolean gdam_source_is_ready_to_read(GdamSource* source)
{
	GtkObjectClass* object_class = GTK_OBJECT(source)->klass;
	GdamSourceClass* klass = GDAM_SOURCE_CLASS(object_class);
	if (klass->is_ready_to_read == NULL)
		return TRUE;
	return klass->is_ready_to_read(source);
}

gboolean gdam_source_detach_child(GdamSource* parent, GdamSource* child)
{
	GtkObjectClass* object_class = GTK_OBJECT(parent)->klass;
	GdamSourceClass* source_class = GDAM_SOURCE_CLASS(object_class);
	return source_class->detach_child(parent, child);
}
void gdam_source_orphan(GdamSource* source)
{
	gtk_object_ref(GTK_OBJECT(source));
	if (source->parent) {
		GdamSource* parent = source->parent;
		if (!gdam_source_detach_child(parent, source))
			;
	} else if (source->device) {
		GdamSink* device = source->device;
		source->device = NULL;
		gdam_sink_remove_source(device);
	}
}

static void add_latency(GdamLatencyInfo* total, GdamLatencyInfo* part)
{
	total->latency = (int) (((gint64)part->resample * total->latency)
			                 / GdamFixed_ONE);
	total->resample = (int) (((gint64)part->resample * total->resample)
			                 / GdamFixed_ONE);
	total->latency += part->latency;
}
	                                 

void gdam_source_find_total_latency(GdamSource* source,
			            GdamLatencyInfo* total)
{
	GdamSource* at = source->parent;
	GdamSource* last = source;
	GdamLatencyInfo latency;
	total->resample = GdamFixed_ONE;
	total->latency = 0;
	while (at) {
		GdamSourceClass* klass;
		klass = GDAM_SOURCE_CLASS(GTK_OBJECT(at)->klass);
		(*klass->get_latency_info)(at, last, &latency);
		add_latency(total, &latency);
		last = at;
		at = at->parent;
	}
}

/* Given the inputs to synchronize (periodb16, offset in samples),
 * returns where in the beat we are. */
static int find_last_beat_offset(GdamTime position,
                                 gint64   period_fixed,
                                 GdamTime offset)
{
	int beat_offset, period_int;
	position -= offset;
	position <<= GdamFixed_LOG_ONE;
	beat_offset = (int) ((position % period_fixed) >> GdamFixed_LOG_ONE);
	if (beat_offset < 0) {
		period_int = (int)(period_fixed >> GdamFixed_LOG_ONE);
		while (beat_offset < 0)
			beat_offset += period_int;
	}
	return beat_offset;
}


#define ABSOLUTE_INTEGER(a)   ( ((a) < 0) ? -(a) : (a) )

gint gdam_source_get_sync_seek(GdamSource*      mobile,
                               GdamSource*      fixed,
			       GdamTime         mobile_period,
			       GdamTime         mobile_offset,
			       GdamTime         fixed_period,
			       GdamTime         fixed_offset,
			       GdamSource*      mobile_latency_proxy)
{
	GdamTime mobile_position, fixed_position;
	int prev_beat_fixed, next_beat_fixed;
	int prev_beat_mobile, next_beat_mobile;
	int toplevel_nudge;
	GdamLatencyInfo total_fixed_latency, total_mobile_latency;

	if (mobile_latency_proxy != NULL)
		gdam_source_find_total_latency(mobile_latency_proxy,
		                   &total_mobile_latency);
	else
		gdam_source_find_total_latency(mobile, &total_mobile_latency);
	gdam_source_find_total_latency(fixed, &total_fixed_latency);

#if 0
	gdam_debug("get_sync_seek: mobile type %s", 
			gtk_type_name(GTK_OBJECT(mobile)->klass->type));
	gdam_debug("get_sync_seek: fixed type %s", 
			gtk_type_name(GTK_OBJECT(fixed)->klass->type));

	gdam_debug("get_sync_seek: mobile_period=%.6g seconds, "
			"static_period=%.6g seconds",
			(gdouble)mobile_period / 44100.0
					/ (gdouble)GdamFixed_ONE,
			(gdouble)fixed_period / 44100.0
					/ (gdouble)GdamFixed_ONE);
	gdam_debug("get_sync_seek: total_fixed_latency=%d samples, "
				"total_mobile_latency=%d samples",
				total_fixed_latency.latency,
				total_mobile_latency.latency);
#endif
	/*
	 * Compute the time (in samples) of the next and last 
	 * beats of each source at the sink-most source (usually a mix)
	 * relative to the current position of that source.
	 */
	mobile_position = gdam_source_get_position(mobile);
	fixed_position = gdam_source_get_position(fixed);
	if (mobile_position < 0 || fixed_position < 0) {
		g_warning("syncing: couldn't get sources position");
		return 0;
	}

#if 0
	gdam_debug("mobile_position=%.6g seconds [%lld], "
	                   "static_position=%.6g [%lld], "
	                   "relative_position=%.6g",
				(gdouble)mobile_position / 44100.0,
				mobile_position,
				(gdouble)fixed_position / 44100.0,
				fixed_position,
				((gdouble)(mobile_position-fixed_position)
				   / 44100.0));
#endif

	/* returns output in samples. */
	next_beat_mobile = prev_beat_mobile = -find_last_beat_offset(
							mobile_position,
							mobile_period,
							mobile_offset);
	next_beat_mobile += (int) (mobile_period / GdamFixed_ONE);
	next_beat_fixed = prev_beat_fixed = -find_last_beat_offset(
							fixed_position,
							fixed_period,
							fixed_offset);
	next_beat_fixed += (int) (fixed_period / GdamFixed_ONE);

	/* Convert the beats from in-song-time to toplevel-mix time. */
	prev_beat_mobile = (int) (((gint64)prev_beat_mobile * GdamFixed_ONE
	                                  / total_mobile_latency.resample)
					  + total_mobile_latency.latency);
	next_beat_mobile = (int) (((gint64)next_beat_mobile * GdamFixed_ONE
	                                  / total_mobile_latency.resample)
					  + total_mobile_latency.latency);
	prev_beat_fixed  = (int) (((gint64)prev_beat_fixed * GdamFixed_ONE
	                                  / total_fixed_latency.resample)
					  + total_fixed_latency.latency);
	next_beat_fixed  = (int) (((gint64)next_beat_fixed * GdamFixed_ONE
	                                  / total_fixed_latency.resample)
					  + total_fixed_latency.latency);

	/*
	 * Compute the amount to seek at the top level.
	 */
	{
		int best_seek = prev_beat_mobile - prev_beat_fixed;
		int seek_abs = ABSOLUTE_INTEGER(best_seek);
		int try_seek, abs_try_seek;

		try_seek = prev_beat_mobile - next_beat_fixed;
		abs_try_seek = ABSOLUTE_INTEGER(try_seek);
		if (abs_try_seek < seek_abs) {
			seek_abs = abs_try_seek;
			best_seek = try_seek;
		}
		try_seek = next_beat_mobile - prev_beat_fixed;
		abs_try_seek = ABSOLUTE_INTEGER(try_seek);
		if (abs_try_seek < seek_abs) {
			seek_abs = abs_try_seek;
			best_seek = try_seek;
		}
		try_seek = next_beat_mobile - next_beat_fixed;
		abs_try_seek = ABSOLUTE_INTEGER(try_seek);
		if (abs_try_seek < seek_abs) {
			seek_abs = abs_try_seek;
			best_seek = try_seek;
		}
		toplevel_nudge = best_seek;

		gdam_debug("prev/next fixed: %d/%d;  prev/next mobile: %d/%d"
		           "; toplevel_nudge=%d",
			prev_beat_fixed, next_beat_fixed,
			prev_beat_mobile, next_beat_mobile,
			toplevel_nudge);
	}

	/*
	 * Translate that nudge to an underlying position.
	 */
	return (int) ((gint64)toplevel_nudge
					* total_mobile_latency.resample
					/ GdamFixed_ONE);
}

static GdamBuffer* gdam_source_real_interpret_command(GdamSource* source,
	  GdamBuffer* command, gboolean* was_ok)
{
	int type;
	*was_ok = FALSE;
	if (!gdam_buffer_peek_int(command, &type)) {
		gdam_debug("gdam_source_real_interpret_command: was invoked, "
			   "and no message...");
		return NULL;
	}
	if (type == GDAM_CFG_SOURCE_SEEK_RELATIVE) {
		GdamTime time;
		gdam_buffer_parse_int(command, &type);
		if (!gdam_buffer_parse_time(command, &time)) {
			*was_ok = FALSE;
			return NULL;
		}
		gdam_dispatch_log("seek relative %lld", time);
		gdam_source_seek(source, time, 1);
		*was_ok = TRUE;
		return NULL;
	}
	if (type == GDAM_CFG_GET_TIME) {
		GdamBuffer* rv = gdam_buffer_new();
		GdamTime position = gdam_source_get_position(source);
		if (position < 0) {
			*was_ok = FALSE;
			return NULL;
		}
		*was_ok = TRUE;
		gdam_buffer_append_int(rv, GDAM_RESPONSE_TIME_SUCCESS);
		gdam_buffer_append_time(rv, position);
		return rv;
	}
	gdam_debug("gdam_source_real_interpret_command: "
			"was invoked... bad command type %d", type);
	return NULL;
}

static GdamTime gdam_source_real_get_position(GdamSource* source)
{
	return -1;
}

GdamTime gdam_source_get_position(GdamSource* source)
{
	GdamSourceClass* source_class;
	source_class = GDAM_SOURCE_CLASS(GTK_OBJECT(source)->klass);
	return (*source_class->get_position)(source);
}


static void gdam_source_ended(GdamSource* source)
{
	source->ended = 1;
}
static void signal_ended(GdamConnection* user,
                         GtkObject* source_object,
			 int connection_id,
                         gpointer unused1,
			 gpointer unused2)
{
	GdamBuffer buffer;
	GdamSource* source = GDAM_SOURCE(source_object);
	g_return_if_fail(source != NULL);
	gdam_buffer_construct(&buffer);
	gdam_buffer_append_int(&buffer, GDAM_EVENT_SOURCE_ENDED);
	gdam_buffer_append_int(&buffer, connection_id);
	gdam_connection_emit_event(user, &buffer);
	gdam_debug("signal_ended: %s",
	           gtk_type_name(source_object->klass->type));
}

guint gdam_source_get_ended_event()
{
	static guint source_ended_id = 0;
	if (!source_ended_id) {
		source_ended_id = gdam_event_register(GDAM_TYPE_SOURCE,
		                                      "ended",
						      signal_ended,
						      NULL,
						      NULL);
		g_assert(source_ended_id != 0);
	}
	return source_ended_id;
}

static void marshal_timer(GdamConnection*    registered,
                          GtkObject*         actor,
			  int                connection_id,
			  gpointer           event_type_data,
			  gpointer           connection_data)
{
	GdamSource* source = GDAM_SOURCE(actor);
	GdamBuffer event;
	gdam_buffer_construct(&event);
	gdam_buffer_append_int(&event, GDAM_EVENT_SOURCE_TIMER);
	gdam_buffer_append_int(&event, connection_id);
	gdam_buffer_append_time(&event, gdam_source_get_position(source));
	gdam_connection_emit_event(registered, &event);
}

guint gdam_source_event_timer()
{
	static guint id = 0;
	if (id == 0) {
		id = gdam_timer_register(GDAM_TYPE_SOURCE,
		                         marshal_timer,
					 NULL,
					 NULL);
	}
	return id;
}

GdamBuffer* gdam_source_interpret_command(GdamSource*     source,
	                                  GdamBuffer*     command,
					  gboolean*       was_success)
{
	GtkObjectClass* klass = ((GtkObject*)source)->klass;
	GdamSourceClass* source_class = (GdamSourceClass*)klass;
	return source_class->interpret_command(source, command, was_success);
}

gboolean gdam_source_seek(GdamSource* source, GdamTime offset, int whence)
{
	GdamSourceClass* source_class;
	source_class = (GdamSourceClass*) ((GtkObject*)source)->klass;
	if (!source_class->seek) {
		g_warning("tried to seek source which didn't have a seek "
		          "method");
		return -1;
	}

	return source_class->seek(source, offset, whence);
}
gboolean gdam_source_is_ancestor(GdamSource* child, GdamSource* potential_par)
{
	while (child) {
		if (child == potential_par)
			return TRUE;
		child = child->parent;
	}
	return FALSE;
}
static int gdam_source_done_constructing_defer_count = 0;
static GSList* pending_done_constructing = NULL;

void gdam_source_done_constructing(GdamSource* source)
{
	GdamSourceClass* source_class;
	source_class = GDAM_SOURCE_CLASS(GTK_OBJECT(source)->klass);
	if (source_class->done_constructing == NULL)
		return;

	if (gdam_source_done_constructing_defer_count > 0)
		pending_done_constructing
			= g_slist_prepend(pending_done_constructing, source);
	else
		(*source_class->done_constructing)(source);
}

void gdam_source_defer_done_constructing()
{
	gdam_source_done_constructing_defer_count++;
}
void gdam_source_undefer_done_constructing()
{
	g_assert(gdam_source_done_constructing_defer_count > 0);
	if (--gdam_source_done_constructing_defer_count == 0) {
		GSList* pending = g_slist_reverse(pending_done_constructing);
		pending_done_constructing = NULL;
		g_slist_foreach(pending,
		                (GFunc) gdam_source_done_constructing,
				NULL);
		g_slist_free(pending);
	}
}

GdamSource* gdam_source_new_generic_v(const char* type_name,
                                      int         num_args,
			              GdamArg*    args)
{
	GtkType gtk_type;
	GdamSource* source;
	if (!type_name) {
		gdam_debug("to gtktype for `%s'", type_name);
		return NULL;
	}
	gtk_type = gdam_plugin_get_type_by_name("server", type_name);
	if (gtk_type == 0) {
		g_warning("no plugin of type %s found", type_name);
		return NULL;
	}
	source = GDAM_SOURCE(gtk_type_new(gtk_type));
	if (!source) {
		gdam_debug("error instantiating source of type %s", type_name);
		return NULL;
	}
	if (num_args > 0) {
		int i;
		for (i = 0; i < num_args; i++)
			gdam_arg_set_object_arg((GtkObject*)source, args + i);
	}
	gdam_source_done_constructing(source);
	return source;
}

void               gdam_source_get_subsources   (GdamSource*     source, 
                                                 int*            num_subs_out,
						 GdamSource***   subs_out)
{
	GtkObjectClass* object_class = GTK_OBJECT(source)->klass;
	GdamSourceClass* source_class = GDAM_SOURCE_CLASS(object_class);
	*num_subs_out = 0;
	*subs_out = NULL;
	if (source_class->get_subsources)
		(*source_class->get_subsources)(source, num_subs_out, subs_out);
	else {
		*num_subs_out = 0;
		*subs_out = NULL;
	}
}


